package ru.dbotthepony.mc.otm.block.entity.tech

import it.unimi.dsi.fastutil.ints.IntArrayList
import net.minecraft.core.BlockPos
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.state.BlockState
import net.neoforged.neoforge.capabilities.Capabilities
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kommons.util.value
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.energy
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.energy.ProfiledEnergyStorage
import ru.dbotthepony.mc.otm.capability.energyStoredMattery
import ru.dbotthepony.mc.otm.capability.matteryEnergy
import ru.dbotthepony.mc.otm.capability.maxEnergyStoredMattery
import ru.dbotthepony.mc.otm.capability.transcieveEnergy
import ru.dbotthepony.mc.otm.container.slotted.FilteredContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.SlottedContainer
import ru.dbotthepony.kommons.guava.immutableList
import ru.dbotthepony.mc.otm.util.isNotEmpty
import ru.dbotthepony.mc.otm.util.math.Decimal
import ru.dbotthepony.mc.otm.util.otmRandom
import ru.dbotthepony.mc.otm.util.shuffle
import ru.dbotthepony.mc.otm.menu.tech.BatteryBankMenu
import ru.dbotthepony.mc.otm.registry.game.MBlockEntities
import java.util.function.Supplier

class BatteryBankBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.BATTERY_BANK, p_155229_, p_155230_), IMatteryEnergyStorage {
	var gaugeLevel by syncher.float()
		private set

	private var containerSlotIndices = IntArray(0)

	private inner class Slot(container: SlottedContainer, slot: Int) : FilteredContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			return super.canAutomationPlaceItem(itemStack) && itemStack.getCapability(Capabilities.EnergyStorage.ITEM)?.let { it.canExtract() && it.extractEnergy(Int.MAX_VALUE, true) > 0 } == true
		}

		override fun canAutomationTakeItem(desired: Int): Boolean {
			return super.canAutomationTakeItem(desired) && item.getCapability(Capabilities.EnergyStorage.ITEM)?.let { !it.canExtract() || it.extractEnergy(Int.MAX_VALUE, true) <= 0 } != false
		}

		override fun notifyChanged(old: ItemStack) {
			super.notifyChanged(old)

			batteryStatus[slot].value = item.getCapability(Capabilities.EnergyStorage.ITEM) != null
			updateGaugeLevel()
		}

		override val maxStackSize: Int
			get() = 1
	}

	private fun containerUpdated() {
		markDirtyFast()

		val newSlots = IntArrayList()

		for (i in 0 until container.containerSize) {
			val stack = container[i]

			if (stack.isNotEmpty && stack.energy != null)
				newSlots.add(i)
		}

		containerSlotIndices = newSlots.toIntArray()
	}

	// 6 на 2
	val container = SlottedContainer.simple(CAPACITY, ::Slot, ::containerUpdated).also(::addDroppableContainer)
	val batteryStatus = immutableList(CAPACITY) { syncher.boolean(false) }

	val itemConfig = ConfigurableItemHandler(inputOutput = container)

	init {
		savetables.stateful(::container, INVENTORY_KEY)
	}

	private fun updateGaugeLevel() {
		var stored = Decimal.ZERO
		var maxStored = Decimal.ZERO

		for (stack in container) {
			val cap = stack.energy ?: continue
			stored += cap.energyStoredMattery
			maxStored += cap.maxEnergyStoredMattery
		}

		gaugeLevel = stored.percentage(maxStored)
	}

	override fun tick() {
		super.tick()
		updateGaugeLevel()
	}

	private fun distributeEnergy(isReceiving: Boolean, howMuch: Decimal, simulate: Boolean): Decimal {
		if (!howMuch.isPositive || containerSlotIndices.isEmpty())
			return Decimal.ZERO

		containerSlotIndices.shuffle(level!!.otmRandom)
		var summ = Decimal.ZERO
		var remaining = howMuch

		for (currentSlot in containerSlotIndices) {
			val item = container[currentSlot]

			if (!item.isEmpty) {
				val energy = item.matteryEnergy ?: item.energy

				if (energy != null) {
					val diff = energy.transcieveEnergy(remaining, isReceiving, simulate)
					summ += diff
					remaining -= diff

					if (!remaining.isPositive) {
						break
					}
				}
			}
		}

		if (!simulate && !summ.isZero) {
			markDirtyFast()
		}

		return summ
	}

	override fun extractEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
		return distributeEnergy(isReceiving = false, howMuch, simulate)
	}

	override fun receiveEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
		return distributeEnergy(isReceiving = true, howMuch, simulate)
	}

	override val canSetBatteryLevel: Boolean
		get() = false

	override val energyFlow: FlowDirection
		get() = FlowDirection.BI_DIRECTIONAL

	override var batteryLevel: Decimal
		get() {
			var result = Decimal.ZERO

			for (stack in container) {
				stack.energy?.let {
					result += it.energyStoredMattery
				}
			}

			return result
		}
		set(value) {
			throw UnsupportedOperationException("Can't set battery value for battery bank")
		}

	override val maxBatteryLevel: Decimal
		get() {
			var result = Decimal.ZERO

			for (stack in container) {
				stack.energy?.let {
					result += it.maxEnergyStoredMattery
				}
			}

			return result
		}

	val energyConfig = ConfigurableEnergy(
		ProfiledEnergyStorage(this),
		frontDefault = FlowDirection.OUTPUT,
		backDefault = FlowDirection.INPUT,
		leftDefault = FlowDirection.INPUT,
		rightDefault = FlowDirection.INPUT,
		topDefault = FlowDirection.INPUT,
		bottomDefault = FlowDirection.INPUT,
		volatileEnergyValues = true
	)

	init {
		energyConfig.front.automatePush = true
		savetables.stateful(Supplier { energyConfig.energy.savedata }, "energyUsageHistory")
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return BatteryBankMenu(containerID, inventory, this)
	}

	companion object {
		const val CAPACITY = 6 * 2
	}
}
