package ru.dbotthepony.mc.otm.player

import net.minecraft.nbt.CompoundTag
import net.minecraft.world.Difficulty
import net.minecraft.world.damagesource.DamageSource
import net.minecraft.world.effect.MobEffectInstance
import net.minecraft.world.effect.MobEffects
import net.minecraft.world.entity.player.Player
import net.minecraft.world.food.FoodConstants
import net.minecraft.world.food.FoodData
import net.minecraft.world.food.FoodProperties
import net.minecraft.world.level.GameRules
import ru.dbotthepony.mc.otm.config.IFoodRegenerationValues
import ru.dbotthepony.mc.otm.config.PlayerConfig
import ru.dbotthepony.mc.otm.util.damageType
import ru.dbotthepony.mc.otm.util.math.Decimal
import ru.dbotthepony.mc.otm.util.getDecimal
import ru.dbotthepony.mc.otm.util.set
import ru.dbotthepony.mc.otm.registry.MDamageTypes
import kotlin.math.max
import kotlin.math.min
import kotlin.math.roundToInt

class MatteryFoodData(private var player: Player) : FoodData() {
	private fun add(foodLevel: Int, saturation: Float) {
		if (player.matteryPlayer.isAndroid && PlayerConfig.REGENERATE_ENERGY) {
			energyToDrain -= PlayerConfig.ANDROID_ENERGY_PER_HUNGER_POINT * foodLevel * PlayerConfig.REGENERATE_ENERGY_EFFICIENCY_FOOD
			energyToDrain -= PlayerConfig.ANDROID_ENERGY_PER_HUNGER_POINT * saturation.coerceAtMost(PlayerConfig.Food.SOFT_FOOD_LIMIT.toFloat()) * PlayerConfig.REGENERATE_ENERGY_EFFICIENCY_SATURATION
		} else if (!player.matteryPlayer.isAndroid) {
			this.foodLevel = min(this.foodLevel + foodLevel, PlayerConfig.Food.HARD_FOOD_LIMIT)
			this.saturationLevel = min(this.saturationLevel + saturation, this.foodLevel + PlayerConfig.Food.OVERSATURATION_LIMIT.toFloat()).coerceAtLeast(0f)
		}
	}

	override fun eat(foodLevelModifier: Int, saturationLevelModifier: Float) {
		add(foodLevelModifier, FoodConstants.saturationByModifier(foodLevelModifier, saturationLevelModifier))
	}

	override fun eat(foodProperties: FoodProperties) {
		add(foodProperties.nutrition(), foodProperties.saturation())
	}

	override fun needsFood(): Boolean {
		return player.matteryPlayer.isAndroid || foodLevel < PlayerConfig.Food.SOFT_FOOD_LIMIT
	}

	override fun addExhaustion(exhaustion: Float) {
		// store exhaustion as usual, and handle its reduction in tick for both humans and androids
		this.exhaustionLevel = min(this.exhaustionLevel + exhaustion, PlayerConfig.Food.EXHAUSTION_LIMIT.toFloat()).coerceAtLeast(0f)
	}

	override fun readAdditionalSaveData(compoundTag: CompoundTag) {
		super.readAdditionalSaveData(compoundTag)

		if ("lastFoodLevel" in compoundTag)
			lastFoodLevel = compoundTag.getInt("lastFoodLevel")

		if ("foodAndroidEnergyToDrain" in compoundTag)
			energyToDrain = compoundTag.getDecimal("foodAndroidEnergyToDrain")
	}

	override fun addAdditionalSaveData(compoundTag: CompoundTag) {
		super.addAdditionalSaveData(compoundTag)
		compoundTag["lastFoodLevel"] = lastFoodLevel
		compoundTag["foodAndroidEnergyToDrain"] = energyToDrain
	}

	private var energyToDrain = Decimal.ZERO

	private fun tickExhaustion() {
		if (player.matteryPlayer.isAndroid && exhaustionLevel > 0f) {
			energyToDrain += PlayerConfig.ANDROID_ENERGY_PER_HUNGER_POINT * (exhaustionLevel / EXHAUSTION_PER_HUNGER_POINT)
			exhaustionLevel = 0f
		} else if (exhaustionLevel >= EXHAUSTION_PER_HUNGER_POINT) {
			var points = (exhaustionLevel / EXHAUSTION_PER_HUNGER_POINT).toInt()
			exhaustionLevel %= EXHAUSTION_PER_HUNGER_POINT

			if (saturationLevel > 0f) {
				val satisfied = min(saturationLevel.roundToInt(), points)
				points -= satisfied
				saturationLevel = max(0f, saturationLevel - satisfied)
			}

			foodLevel = max(0, foodLevel - points)
		}
	}

	private fun tickRegeneration(values: IFoodRegenerationValues): Boolean {
		if (!player.isHurt || foodLevel < values.foodLimit || values.requiresSaturation && saturationLevel <= 0f)
			return false

		while (player.isHurt && foodLevel >= values.foodLimit && (!values.requiresSaturation || saturationLevel > 0f) && ++tickTimer >= values.ticks) {
			tickTimer = 0

			val healAmount = if (values.requiresSaturation && values.regenerationSlowdown) min(saturationLevel, values.upperSlowdownBound.toFloat()) / values.upperSlowdownBound.toFloat() else 1f
			player.heal(healAmount)
			addExhaustion(healAmount * 6f)
			tickExhaustion()
		}

		return tickTimer != 0
	}

	override fun tick(player: Player) {
		this.player = player

		tickExhaustion()

		if (player.matteryPlayer.isAndroid) {
			if (energyToDrain > Decimal.ZERO)
				energyToDrain -= player.matteryPlayer.androidEnergy.extractEnergy(energyToDrain, false)
			else if (energyToDrain < Decimal.ZERO) {
				energyToDrain += player.matteryPlayer.androidEnergy.receiveEnergy(-energyToDrain, false)

				if (player.matteryPlayer.androidEnergy.missingPower <= Decimal.ZERO)
					energyToDrain = Decimal.ZERO
			}

			if (player.level().difficulty == Difficulty.PEACEFUL && PlayerConfig.REGENERATE_ENERGY_IN_PEACEFUL)
				player.matteryPlayer.androidEnergy.receiveEnergy(PlayerConfig.ANDROID_ENERGY_PER_HUNGER_POINT, false)

			if (!player.matteryPlayer.androidHasEnergy) {
				if (++tickTimer >= 20 && player.hurt(DamageSource(player.level().registryAccess().damageType(MDamageTypes.ANDROID_DISCHARGE)), 1f)) {
					tickTimer = 0
				}

				val effect = player.activeEffectsMap[MobEffects.MOVEMENT_SLOWDOWN]

				if (effect == null || effect.duration < 40 || effect.amplifier < 2) {
					player.addEffect(MobEffectInstance(MobEffects.MOVEMENT_SLOWDOWN, effect?.duration?.coerceAtLeast(60) ?: 60, effect?.amplifier?.coerceAtLeast(2) ?: 2, false, false))
				}
			} else {
				if (player.isHurt && player.level().gameRules.getBoolean(GameRules.RULE_NATURAL_REGENERATION)) {
					if (++tickTimer >= PlayerConfig.TIME_BETWEEN_NATURAL_REGENERATION) {
						player.heal(1f)
						addExhaustion(6f)
						tickTimer = 0
					}
				} else {
					tickTimer = 0
				}
			}
		} else {
			lastFoodLevel = foodLevel
			energyToDrain = Decimal.ZERO

			if (
				player.level().gameRules.getBoolean(GameRules.RULE_NATURAL_REGENERATION) &&
				(tickRegeneration(PlayerConfig.Food.FAST_REGEN) || tickRegeneration(PlayerConfig.Food.SLOW_REGEN))
			) {
				// do nothing
			} else if (PlayerConfig.Food.ENABLE_STARVATION && foodLevel <= 0) {
				if (++tickTimer >= PlayerConfig.Food.STARVATION_TICKS) {
					tickTimer = 0

					val threshold = when (player.level().difficulty) {
						Difficulty.PEACEFUL -> Float.POSITIVE_INFINITY
						Difficulty.EASY -> PlayerConfig.Food.STARVATION_HEALTH_LIMIT_EASY.toFloat()
						Difficulty.NORMAL -> PlayerConfig.Food.STARVATION_HEALTH_LIMIT_NORMAL.toFloat()
						Difficulty.HARD -> PlayerConfig.Food.STARVATION_HEALTH_LIMIT_HARD.toFloat()
					}

					if (player.health > threshold) {
						player.hurt(player.damageSources().starve(), 1.0f)
					}
				}
			} else {
				tickTimer = 0
			}
		}
	}

	override fun getFoodLevel(): Int {
		if (player.matteryPlayer.isAndroid)
			return if (player.matteryPlayer.androidHasEnergy) PlayerConfig.Food.SOFT_FOOD_LIMIT else 0

		return super.getFoodLevel()
	}

	override fun getLastFoodLevel(): Int {
		if (player.matteryPlayer.isAndroid)
			return getFoodLevel()

		return super.getLastFoodLevel()
	}

	override fun getExhaustionLevel(): Float {
		if (player.matteryPlayer.isAndroid && player.matteryPlayer.androidHasEnergy)
			return 0f

		return super.getExhaustionLevel()
	}

	override fun getSaturationLevel(): Float {
		if (player.matteryPlayer.isAndroid)
			return if (player.matteryPlayer.androidHasEnergy) player.matteryPlayer.androidEnergy.batteryLevel.percentage(player.matteryPlayer.androidEnergy.maxBatteryLevel) * PlayerConfig.Food.SOFT_FOOD_LIMIT else 0f

		return super.getSaturationLevel()
	}

	companion object {
		const val EXHAUSTION_PER_HUNGER_POINT = 4f
	}
}
