package ru.dbotthepony.mc.otm.server.world

import com.github.benmanes.caffeine.cache.Cache
import com.github.benmanes.caffeine.cache.Caffeine
import com.github.benmanes.caffeine.cache.Scheduler
import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ObjectRBTreeSet
import net.minecraft.Util
import net.minecraft.core.BlockPos
import net.minecraft.core.Holder
import net.minecraft.core.SectionPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.WorldGenLevel
import net.minecraft.world.level.levelgen.feature.ConfiguredFeature
import net.minecraft.world.level.levelgen.feature.Feature
import net.minecraft.world.level.levelgen.feature.FeaturePlaceContext
import net.minecraft.world.level.levelgen.feature.configurations.FeatureConfiguration
import net.minecraft.world.level.levelgen.placement.PlacedFeature
import ru.dbotthepony.kommons.util.XXHash64
import ru.dbotthepony.mc.otm.THREAD_LOCAL_RANDOM
import ru.dbotthepony.mc.otm.data.codec.minRange
import ru.dbotthepony.mc.otm.util.GJRAND64RandomSource
import ru.dbotthepony.mc.otm.util.shuffle
import ru.dbotthepony.mc.otm.server.world.feature.EnhancedFeature
import ru.dbotthepony.mc.otm.server.world.placement.EnhancedPlacement
import java.io.DataOutputStream
import java.time.Duration
import java.util.*
import kotlin.collections.ArrayList
import kotlin.math.sqrt

object EnhancedPlacedFeature : Feature<EnhancedPlacedFeature.Config>(
	RecordCodecBuilder.create {
		it.group(
			Codec.INT.minRange(0).fieldOf("chunk_scan_range").forGetter(Config::chunkScanRange),
			Codec.LONG.fieldOf("seed_mix").forGetter(Config::seedMix),
			EnhancedPlacement.CODEC.fieldOf("root").forGetter(Config::root),
		).apply(it, ::Config)
	}
) {
	private val emptyVariableMap = PlacementVariableMap()

	fun configure(
		chunkScanRange: Int,
		seedMix: Long,
		root: EnhancedPlacement
	): PlacedFeature {
		return PlacedFeature(Holder.direct(ConfiguredFeature(EnhancedPlacedFeature, Config(chunkScanRange, seedMix, root))), listOf())
	}

	class Config(
		val chunkScanRange: Int,
		val seedMix: Long,
		val root: EnhancedPlacement,
	) : FeatureConfiguration {
		private class GeneratedChunk : EnhancedPlacementContext.Placer {
			private data class Placement(val context: EnhancedPlacementContext, val positions: TreeSet<PlacementPos>, val feature: EnhancedFeature.Configured<*, *>)
			private val placed = LinkedList<Placement>()

			override fun place(context: EnhancedPlacementContext, positions: List<PlacementPos>, feature: EnhancedFeature.Configured<*, *>) {
				if (positions.isNotEmpty()) {
					placed.add(Placement(context, TreeSet<PlacementPos>().also { it.addAll(positions) }, feature))
				}
			}

			fun place(context: FeaturePlaceContext<*>): Boolean {
				var any = false
				val pos = ChunkPos(context.origin())

				for ((eContext, positions, feature) in placed) {
					val itr = positions.tailSet(PlacementPos(BlockPos(pos.minBlockX, Int.MIN_VALUE, pos.minBlockZ), emptyVariableMap)).iterator()

					if (!itr.hasNext())
						continue

					val result = TreeSet<PlacementPos>()

					for (placementPos in itr) {
						if (SectionPos.blockToSectionCoord(placementPos.pos.x) != pos.x || SectionPos.blockToSectionCoord(placementPos.pos.z) != pos.z)
							break

						result.add(placementPos)
					}

					if (result.isNotEmpty()) {
						any = feature.place(eContext.push(context.level()), result, positions) || any
					}
				}

				return any
			}
		}

		private val level2cache = Caffeine.newBuilder()
			.scheduler(Scheduler.systemScheduler())
			.executor(Util.backgroundExecutor())
			.expireAfterAccess(Duration.ofMinutes(10))
			.weakKeys()
			.build<ServerLevel, Cache<ChunkPos, GeneratedChunk>>()

		private fun getCache(level: WorldGenLevel): Cache<ChunkPos, GeneratedChunk> {
			return level2cache.get(level.level) {
				Caffeine.newBuilder()
					.scheduler(Scheduler.systemScheduler())
					.executor(Util.backgroundExecutor())
					.maximumSize(16384L)
					.expireAfterWrite(Duration.ofMinutes(5))
					.softValues()
					.build()
			}
		}

		private fun evaluate(context: FeaturePlaceContext<*>, chunkPos: ChunkPos): GeneratedChunk {
			val bytes = FastByteArrayOutputStream()
			val dataStream = DataOutputStream(bytes)

			dataStream.writeLong(seedMix)
			dataStream.writeInt(chunkPos.x)
			dataStream.writeInt(chunkPos.z)

			val hash = XXHash64()
			hash.update(bytes.array, 0, bytes.length)

			val random = GJRAND64RandomSource(context.level().seed, hash.digestAsLong())

			val chunk = GeneratedChunk()
			val enhancedContext = EnhancedPlacementContext(context.level(), context.chunkGenerator(), random, chunkPos, chunk)
			root.evaluate(enhancedContext, listOf(PlacementPos(BlockPos(chunkPos.minBlockX, 0, chunkPos.minBlockZ), PlacementVariableMap())))
			return chunk
		}

		fun place(context: FeaturePlaceContext<*>): Boolean {
			val cache = getCache(context.level())
			val chunkPos = ChunkPos(context.origin())
			val chunkPositions = ArrayList<ChunkPos>()

			for (x in -chunkScanRange .. chunkScanRange) {
				for (z in -chunkScanRange .. chunkScanRange) {
					// floor, so chunk scan range of 1 will give square instead of diamond
					val radius = sqrt(x.toDouble() * x + z.toDouble() * z).toInt()

					if (radius <= chunkScanRange) {
						val thisPos = ChunkPos(chunkPos.x + x, chunkPos.z + z)
						chunkPositions.add(thisPos)
					}
				}
			}

			val withIndex = ObjectArrayList(chunkPositions.withIndex().iterator())
			// evaluate chunks in random order, so in multithreaded environments we experience less congestion
			// between threads which lookup cache while it is being populated on neighbouring chunks
			withIndex.shuffle(THREAD_LOCAL_RANDOM)

			data class P(val i: Int, val v: GeneratedChunk) : Comparable<P> {
				override fun compareTo(other: P): Int {
					return i.compareTo(other.i)
				}
			}

			val instances = ArrayList<P>()
			instances.sort()

			withIndex.forEach { (i, it) ->
				instances.add(P(i, cache.get(it) { evaluate(context, it) }))
			}

			var any = false
			instances.forEach { any = it.v.place(context) }
			return any
		}
	}

	override fun place(context: FeaturePlaceContext<Config>): Boolean {
		return context.config().place(context)
	}
}
