package ru.dbotthepony.mc.otm.server.world.placement

import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.ConstantFloat
import net.minecraft.util.valueproviders.FloatProvider
import net.minecraft.util.valueproviders.IntProvider
import net.minecraft.world.level.levelgen.placement.PlacementContext
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import net.minecraft.world.level.levelgen.placement.PlacementModifierType
import ru.dbotthepony.mc.otm.util.collect.Vec3iHashStrategy
import ru.dbotthepony.mc.otm.data.world.BooleanProvider
import ru.dbotthepony.mc.otm.registry.data.MPlacementModifiers
import ru.dbotthepony.mc.otm.util.math.VECTOR_POSITIVE_X
import ru.dbotthepony.mc.otm.util.math.VECTOR_POSITIVE_Z
import ru.dbotthepony.mc.otm.util.math.VECTOR_UP
import ru.dbotthepony.mc.otm.util.math.Vector
import ru.dbotthepony.mc.otm.util.math.rotate
import ru.dbotthepony.mc.otm.util.math.rotateAroundThis
import ru.dbotthepony.mc.otm.util.nextDouble
import ru.dbotthepony.mc.otm.server.world.EnhancedPlacementContext
import ru.dbotthepony.mc.otm.server.world.PlacementPos
import java.util.*
import java.util.stream.Stream
import kotlin.math.PI

/**
 * Ellipsoid ("cloud") placement
 */
data class EllipsoidPlacement(
	/**
	 * X position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val x: FloatProvider,

	/**
	 * Z position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val z: FloatProvider,

	/**
	 * Y position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val y: FloatProvider,

	/**
	 * Total amount of block positions to generate
	 */
	val count: IntProvider,

	/**
	 * Ellipsoid size sampler on X axis
	 */
	val xLength: FloatProvider,

	/**
	 * Ellipsoid size sampler on Z axis
	 */
	val zLength: FloatProvider,

	/**
	 * Ellipsoid size sampler on Y axis
	 */
	val yLength: FloatProvider,

	val rotateXZ: BooleanProvider = BooleanProvider.AlwaysFalse,
	val rotateXY: FloatProvider = ConstantFloat.ZERO,
	val rotateZY: FloatProvider = ConstantFloat.ZERO,
) : PlacementModifier(), EnhancedPlacement {
	init {
		require(xLength.minValue >= 1f) { "Bad ellipsoid x minimal size: $xLength" }
		require(zLength.minValue >= 1f) { "Bad ellipsoid z minimal size: $zLength" }
		require(yLength.minValue >= 1f) { "Bad ellipsoid y minimal size: $yLength" }
	}

	private fun evaluate(random: RandomSource, position: BlockPos): Set<BlockPos> {
		val count = count.sample(random)
		val results = TreeSet<BlockPos>(Vec3iHashStrategy)

		if (count <= 0)
			return results

		val xLength = xLength.sample(random)
		val zLength = zLength.sample(random)
		val yLength = yLength.sample(random)

		val xPow = xLength * xLength
		val zPow = zLength * zLength
		val yPow = yLength * yLength

		var iterations = 0
		val maxIterations = count * 10

		val rotateXZ = rotateXZ.instance().sample(random)
		val rotateXY = rotateXY.sample(random)
		val rotateZY = rotateZY.sample(random)

		val rotationXZ: Double

		if (rotateXZ) {
			rotationXZ = random.nextDouble(-PI, PI)
		} else {
			rotationXZ = 0.0
		}

		val qXZ = VECTOR_UP.rotateAroundThis(rotationXZ)
		val qXY = VECTOR_POSITIVE_Z.rotateAroundThis(rotateXY)
		val qZY = VECTOR_POSITIVE_X.rotateAroundThis(rotateZY)

		while (results.size < count && ++iterations < maxIterations) {
			val x = this.x.sample(random) * xLength
			val y = this.y.sample(random) * yLength
			val z = this.z.sample(random) * zLength

			val isValidPoint = (x * x) / xPow +
				(y * y) / yPow +
				(z * z) / zPow <= 1.0f

			if (isValidPoint) {
				var point = Vector(x.toDouble(), y.toDouble(), z.toDouble())

				point = point.rotate(qXZ)
				point = point.rotate(qXY)
				point = point.rotate(qZY)

				results.add(BlockPos(point.x.toInt() + position.x, point.y.toInt() + position.y, point.z.toInt() + position.z))
			}
		}

		return results
	}

	override fun getPositions(
		context: PlacementContext,
		random: RandomSource,
		position: BlockPos
	): Stream<BlockPos> {
		return evaluate(random, position).stream()
	}

	override fun evaluate(context: EnhancedPlacementContext, positions: List<PlacementPos>): List<PlacementPos> {
		val result = ArrayList<PlacementPos>()
		val results = positions.map { (it, c) -> evaluate(context.random, it).map { PlacementPos(it, c) } }
		result.ensureCapacity(result.size + results.sumOf { it.size } + 1)
		results.forEach { result.addAll(it) }
		return result
	}

	override val type: EnhancedPlacement.Type<*>
		get() = Companion

	override fun type(): PlacementModifierType<*> {
		return MPlacementModifiers.ELLIPSOID_PLACEMENT
	}

	companion object : EnhancedPlacement.Type<EllipsoidPlacement> {
		val CODEC: MapCodec<EllipsoidPlacement> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					FloatProvider.CODEC.fieldOf("x").forGetter(EllipsoidPlacement::x),
					FloatProvider.CODEC.fieldOf("y").forGetter(EllipsoidPlacement::y),
					FloatProvider.CODEC.fieldOf("z").forGetter(EllipsoidPlacement::z),
					IntProvider.CODEC.fieldOf("count").forGetter(EllipsoidPlacement::count),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("x_size").forGetter(EllipsoidPlacement::xLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("z_size").forGetter(EllipsoidPlacement::zLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("y_size").forGetter(EllipsoidPlacement::yLength),
					BooleanProvider.CODEC.optionalFieldOf("rotateXZ", BooleanProvider.AlwaysFalse).forGetter(EllipsoidPlacement::rotateXZ),
					FloatProvider.CODEC.optionalFieldOf("rotateXY", ConstantFloat.ZERO).forGetter(EllipsoidPlacement::rotateXY),
					FloatProvider.CODEC.optionalFieldOf("rotateZY", ConstantFloat.ZERO).forGetter(EllipsoidPlacement::rotateZY),
				).apply(it, ::EllipsoidPlacement)
			}
		}

		override val codec: MapCodec<EllipsoidPlacement>
			get() = CODEC
	}
}
